"""Module for searching torrents on TokyoTosho."""
from typing import List, Tuple

import bs4

from mariner import searchengine

Magnet = str
Name = str
Url = str


class TokyoTosho(searchengine.SearchEngine):
    """Represents TokyoTosho search engine."""

    search_url = 'https://www.tokyotosho.info/search.php?terms='

    @staticmethod
    def _parse(raw: str) -> List[Tuple[Name, Magnet, Url]]:
        """Parse result page.

        Args:
          raw: Raw HTML results to parse.

        Returns:
            List of torrent names with magnet links and URLs.
        """
        soup = bs4.BeautifulSoup(raw, 'lxml')
        torrents = soup.select('.desc-top')
        results = []
        for torrent_ in torrents:
            links = torrent_.select('a')
            magnet = links[0].get('href')
            url = links[1].get('href')
            name = TokyoTosho._parse_name(links[1].contents)
            results.append((name, magnet, url))
        return results

    @staticmethod
    def _parse_name(raw: str) -> Name:
        """Parse torrent name.

        Args:
          raw: Raw HTML name to parse.

        Returns:
            Parsed name.
        """
        name = [c for c in raw if isinstance(c, bs4.element.NavigableString)]
        return ''.join(name)



# -*- coding: future_fstrings -*-
"""Module for searching torrents on Distrowatch."""
import asyncio
import logging
from typing import List, Iterable, Tuple

import bs4

from mariner import searchengine, torrent, cache

Url = str
Page = str
Name = str


class Distrowatch(searchengine.SearchEngine):
    """Represent Distrowatch search engine."""

    log = logging.getLogger(__name__)
    search_url = "https://distrowatch.com/dwres.php?resource=bittorrent"

    async def get_results(self, title: str) -> None:
        """Get of list of torrent page urls.

        Args:
            title: String to search for.
        """
        try:
            page = await self.get(self.search_url)
        except OSError:
            print('Cannot reach server')
        else:
            self.urls = self._parse(page)

    @cache.Cache(size=100)
    def get_torrents(self, title: str) -> Iterable[torrent.Torrent]:
        """Get a list of torrents that we searched for.

        Args:
            title: String to search for.

        Returns:
            List of torrent results.
        """
        tid = 0
        torrents = []

        if not self.urls:
            self.log.debug('Fetching results')
            tasks = asyncio.wait([self.get_results(title)])
            loop = asyncio.get_event_loop()
            loop.run_until_complete(tasks)

        for name, url in self.urls:
            if title.lower() in name:
                self.log.debug('Appending tid=%s name=%s', tid, name)
                torrents.append(torrent.Torrent(tid, name, url))
                tid += 1
        return torrents

    @staticmethod
    def _parse(raw: str) -> List[Tuple[Name, Url]]:
        """Parse result page.

        Args:
          raw: Raw HTML page.

        Returns:
            List of torrent names with URLs.
        """
        soup = bs4.BeautifulSoup(raw, 'lxml')
        torrents = soup.find_all('td', 'torrent')
        results = []
        for torrent_ in torrents:
            link = torrent_.find('a')
            name = link.string.lower()
            url_stub = link.get('href')
            url = f"https://distrowatch.com/{url_stub}"
            results.append((name, url))
        return results



# -*- coding: future_fstrings -*-
"""Represent a downloadable torrent."""
from typing import Optional


class Torrent:
    """Class representing a Torrent."""

    def __init__(self,
                 tid: int,
                 name: str,
                 torrent_url: str,
                 *,
                 magnet_link: Optional[str] = None,
                 description: Optional[str] = None) -> None:
        self.tid = tid
        self.name = name
        self.torrent_url = torrent_url
        self.magnet_link = magnet_link
        self.description = description

    @property
    def tid(self) -> int:
        """ID of the torrent."""
        return self.__tid

    @tid.setter
    def tid(self, value: int) -> None:
        if value < 0:
            raise ValueError("ID must be a positive number")
        else:
            self.__tid = value  # pylint: disable=attribute-defined-outside-init

    @property
    def mods(self) -> str:
        """Show whether it is available as torrent, magnet link or both."""
        mods = []
        if self.torrent_url:
            mods.append('Torrent')
        if self.magnet_link:
            mods.append('Magnet link')
        return ', '.join(mods)

    @property
    def filename(self) -> str:
        """Filename to use for the torrent."""
        if '.torrent' in self.name:
            return self.name
        return f'{self.name}.torrent'

    def __repr__(self) -> str:
        return f"Torrent({self.tid}, {self.name}, {self.torrent_url}," \
            "{self.magnet_link}, {self.description})"

    def __str__(self) -> str:
        return self.__repr__()



# -*- coding: future_fstrings -*-
"""Handle searching for torrents on torrent trackers."""
import abc
import asyncio
import importlib
import logging
import pathlib
from typing import List, Iterable, Tuple, Optional, Union

import aiohttp
import async_timeout

from mariner import torrent, cache

Name = str
Page = str
Path = Union[str, pathlib.Path]
Url = str

engines = {}


class SearchEngineManager:
    """Manage search engine plugins."""
    log = logging.getLogger(__name__)

    def __init__(self, path: Path = None) -> None:
        if not path:
            self.engine_directory = pathlib.Path(__file__).parent / 'plugins'
        else:
            self.engine_directory = pathlib.Path(path)
        self.log.debug('path=%s engine_directory=%s',
                       path, self.engine_directory)

    def find_engines(self) -> None:
        """Find and import search engines."""
        for module in self.engine_directory.glob('*.py'):
            self.log.debug('Loading module=%s', module)
            if module.name != '__init__.py':
                name = module.stem
                spec = importlib.util.spec_from_file_location(name, module)
                loaded_module = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(loaded_module)

    def register_engines(self) -> None:
        """Register search engines into global namespace."""
        for engine in SearchEngine.__subclasses__():
            self.log.debug('Adding engine=%s', engine)
            name = engine.__name__.lower()
            engines[name] = engine

    def initialize_engines(self) -> None:
        """Find engines and put them into global namespace."""
        self.log.info('Initializing plugins')
        self.find_engines()
        self.register_engines()


class SearchEngine(abc.ABC):
    """Represent a search engine."""
    log = logging.getLogger(__name__)
    user_agent = {'user-agent': 'Mariner Torrent Downloader'}
    search_url = ''  # To be overwritten by subclasses

    def __init__(self) -> None:
        super().__init__()
        self.results = []
        self.urls = []

    async def get(self, url: Url) -> Page:
        """Asynchronous https request.

        Args:
            url: Url of the webpage to fetch.

        Returns:
            Fetched webpage."""
        self.log.debug('Fetching url=%s', url)
        with async_timeout.timeout(10):
            async with aiohttp.ClientSession() as session:
                async with session.get(url, headers=self.user_agent) as response:
                    page = await response.text()
        return page

    async def get_results(self, title: str) -> None:
        """Get a list of torrent name with URLs and magnet links.

        Args:
            title: String to search for.
        """
        try:
            search_url = self.search_url + title
            page = await self.get(search_url)
        except (OSError, asyncio.TimeoutError):
            self.log.error('Cannot reach server')
        else:
            self.urls = self._parse(page)

    def get_torrent(self, tid: int) -> torrent.Torrent:
        """Get torrent of given id.

        Args:
            tid: ID of the torrents to get.

        Returns:
            Torrent with given ID.

        """
        if not self.results:
            self.log.debug('Fetching results from cache')
            self.results = list(cache.Cache().newest)
        if tid < 0 or tid > len(self.results):
            self.log.debug('tid=%s', tid)
            raise NoResultException(f"No torrent with ID {tid}")
        return self.results[tid]

    @cache.Cache(size=100)
    def get_torrents(self, title: str) -> Iterable[torrent.Torrent]:
        """Get a list of torrents that we searched for.

        Args:
           title: String to search for.

        Returns:
            List of torrent results.
        """
        if not self.urls:
            self.log.debug('Fetching search results')
            tasks = asyncio.wait([self.get_results(title)])
            loop = asyncio.get_event_loop()
            loop.run_until_complete(tasks)

        tid = 0
        torrents = []
        for name, magnet, url in self.urls:
            self.log.debug('Added tid=%s name=%s', tid, name)
            torrents.append(torrent.Torrent(
                tid, name, url, magnet_link=magnet))
            tid += 1
        return torrents

    @staticmethod
    @abc.abstractmethod
    def _parse(raw: str) -> List[Tuple[Name, Url]]:
        """Parse result page.

        Args:
            raw: Raw HTML page.

        Returns:
            List of torrents names with URLs and magnet links.

        """
        raise NotImplementedError

    def search(self, title: str, limit: Optional[int] = 10) -> List[torrent.Torrent]:
        """Search for torrents on given site.

        Args:
            title: String to search for.
            limit: Defaults to 10. Limits the number of results, that should be shown.

        Returns:
            List of Torrents returned by the search, up to the limit.
        """
        if not title:
            raise ValueError('No string to search for.')
        if limit <= 0:
            raise ValueError('Limit has to be higher than zero.')
        torrents = self.get_torrents(title)
        self.results = torrents[:limit]
        self.log.debug('Search results=%s', self.results)
        if self.results:
            return self.results
        raise NoResultException(f"No results for {title}")


class Error(Exception):
    """Base-class for all exceptions raised by this module."""


class NoResultException(Error):
    """No result found for given search string."""



"""Mariner is a command line torrent searcher."""
VERSION = (0, 1, 1)

__title__ = 'mariner'
__description__ = 'Command line torrent searcher'
__url__ = 'http://gitlab.com/radek-sprta/mariner'
__download_url__ = "https://gitlab.com/radek-sprta/mariner/repository/archive.tar.gz?ref=master"
__version__ = '.'.join(map(str, VERSION))
__author__ = 'Radek Sprta'
__author_email__ = 'mail@radeksprta.eu'
__license__ = 'GNU GPLv3'
__copyright__ = 'Copyright 2017 Radek Sprta'
